/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComAdaptiveFilter.cpp
    \brief    adaptive filter class
*/
#include "TComAdaptiveFilter.h"

#if SVC_IL_ADAPTIVE_FILTER

//! \ingroup TLibEncoder
//! \{

// ====================================================================================================================
// Global variables
// ====================================================================================================================

Int* g_AFNumCoef= NULL;

const AFFilter& AFFilter::operator= (const AFFilter& src)
{
  enableFlag  = src.enableFlag;
  filterShape = src.filterShape;
  ::memcpy(coeff, src.coeff, sizeof(Int)*AF_MAX_NUM_COEF);

  return *this;
}

AFFilter::AFFilter()
{
  enableFlag  = false;
  filterShape = 0; 
}


// ====================================================================================================================
// Constructor / destructor / create / destroy / operator
// ====================================================================================================================


TComAdaptiveFilter::TComAdaptiveFilter()
{
  m_tempPicYuv = NULL;
  for(Int f=0; f< AF_NUM_FILTER_SHAPE; f++)
  {
    m_shapeWeight[f] = NULL;
  }
}

TComAdaptiveFilter::~TComAdaptiveFilter()
{
  destroy();
}



Void TComAdaptiveFilter::create( Int picWidth, Int picHeight, UInt maxCUWidth, UInt maxCUHeight, UInt maxCUDepth )
{
  m_picWidth = picWidth;    
  m_picHeight= picHeight;
  m_maxCUWidth= maxCUWidth; 
  m_maxCUHeight= maxCUHeight;
  m_maxCUDepth = maxCUDepth;

  m_numCTUInWidth = (m_picWidth/m_maxCUWidth) + ((m_picWidth % m_maxCUWidth)?1:0);
  m_numCTUInHeight= (m_picHeight/m_maxCUHeight) + ((m_picHeight % m_maxCUHeight)?1:0);

  if ( !m_tempPicYuv )
  {
    m_tempPicYuv = new TComPicYuv;
    m_tempPicYuv->create( m_picWidth, m_picHeight, m_maxCUWidth, m_maxCUHeight, m_maxCUDepth );
  }

  createFilterShapeInfo();
}

Void TComAdaptiveFilter::destroy()
{
  if ( m_tempPicYuv )
  {
    m_tempPicYuv->destroy();
    delete m_tempPicYuv;
    m_tempPicYuv = NULL;
  }
  destroyFilterShapeInfo();
}

Void TComAdaptiveFilter::createFilterShapeInfo()
{
  g_AFNumCoef= new Int[AF_NUM_FILTER_SHAPE];

  for(Int shape=0; shape< AF_NUM_FILTER_SHAPE; shape++)
  {
    switch(shape)
    {
    case AF_SHAPE_7x7CROSS_3x3SQUARE:
      {
        g_AFNumCoef[shape] = 9;

      }
      break;
      /*
      case AF....
      
      */
    default:
      {
        printf("Not a supported shape\n");
        assert(0);
        exit(-1);
      }
    }
  }

  for(Int shape=0; shape< AF_NUM_FILTER_SHAPE; shape++)
  {
    //weight
    m_shapeWeight[shape] = new Int[g_AFNumCoef[shape]];
    for(Int i=0; i< g_AFNumCoef[shape]-1; i++)
    {
      m_shapeWeight[shape][i] = 2;
    }
    m_shapeWeight[shape][g_AFNumCoef[shape]-1] = 1;



  }

}

Void TComAdaptiveFilter::destroyFilterShapeInfo()
{
  if(g_AFNumCoef)
  {
    delete[] g_AFNumCoef;
    g_AFNumCoef = NULL;
  }
  for(Int f=0; f< AF_NUM_FILTER_SHAPE; f++)
  {
    if(m_shapeWeight[f])
    {
      delete[] m_shapeWeight[f]; m_shapeWeight[f] = NULL;
    }
  }

}

// ====================================================================================================================
// methods
// ====================================================================================================================

Void TComAdaptiveFilter::filterBlock(Int shape, Int* coef, Pel* srcBlk, Pel* resBlk, Int srcStride, Int resStride,  Int width, Int height)
{
  Int  offset = (1<<( (Int)AF_NUM_BIT_SHIFT-1));
  Pel* imgPad = srcBlk;
  Pel* imgRes = resBlk;
  Int i, j, pixelInt;

  switch(shape)
  {
  case AF_SHAPE_7x7CROSS_3x3SQUARE:
    {
      Pel *imgPad1,*imgPad2,*imgPad3, *imgPad4, *imgPad5, *imgPad6;
      for(i= 0; i< height; i++)
      {
        imgPad1 = imgPad +   srcStride;
        imgPad2 = imgPad -   srcStride;
        imgPad3 = imgPad + 2*srcStride;
        imgPad4 = imgPad - 2*srcStride;
        imgPad5 = imgPad + 3*srcStride;
        imgPad6 = imgPad - 3*srcStride;
        for(j= 0; j< width ; j++)
        {
          pixelInt  = coef[0]* (imgPad5[j  ] + imgPad6[j  ]);

          pixelInt += coef[1]* (imgPad3[j  ] + imgPad4[j  ]);

          pixelInt += coef[2]* (imgPad1[j+1] + imgPad2[j-1]);
          pixelInt += coef[3]* (imgPad1[j  ] + imgPad2[j  ]);
          pixelInt += coef[4]* (imgPad1[j-1] + imgPad2[j+1]);

          pixelInt += coef[5]* (imgPad[j+3] + imgPad[j-3]);
          pixelInt += coef[6]* (imgPad[j+2] + imgPad[j-2]);
          pixelInt += coef[7]* (imgPad[j+1] + imgPad[j-1]);
          pixelInt += coef[8]* (imgPad[j  ]);

          pixelInt=(Int)((pixelInt+offset) >> AF_NUM_BIT_SHIFT);
          imgRes[j] = Clip( pixelInt );
        }
        imgPad += srcStride;
        imgRes += resStride;
      }
    }
    break;
  default:
    {
      printf("Not a supported filter shape\n");
      assert(0);
      exit(-1);
    }
  }
}

Pel* TComAdaptiveFilter::getPicBuf(TComPicYuv* pPicYuv, Int compIdx)
{
  Pel* pBuf = NULL;
  switch(compIdx)
  {
  case AF_Y:
    {
      pBuf = pPicYuv->getLumaAddr();
    }
    break;
  case AF_Cb:
    {
      pBuf = pPicYuv->getCbAddr();
    }
    break;
  case AF_Cr:
    {
      pBuf = pPicYuv->getCrAddr();
    }
    break;
  default:
    {
      printf("Not a legal component ID for ALF\n");
      assert(0);
      exit(-1);
    }
  }

  return pBuf;
}

Void TComAdaptiveFilter::checkFilterCoeffValue( Int *coef, Int filterLength)
{
  Int maxValueNonCenter = 1 * (1 << AF_NUM_BIT_SHIFT) - 1;
  Int minValueNonCenter = 0 - 1 * (1 << AF_NUM_BIT_SHIFT);
  Int maxValueCenter    = 2 * (1 << AF_NUM_BIT_SHIFT) - 1;
  Int minValueCenter    = 0 ; 

  for(Int i = 0; i < filterLength-1; i++)
  {
    coef[i] = Clip3(minValueNonCenter, maxValueNonCenter, coef[i]);
  }

  coef[filterLength-1] = Clip3(minValueCenter, maxValueCenter, coef[filterLength-1]);
}

Void TComAdaptiveFilter::reconstructCoef(Int* dstCoef, Int* srcCoef, Int shape/* =0 */)
{
  Int numCoef = g_AFNumCoef[shape];
  Int sum=0;

  for(Int i=0; i< numCoef-1;i++)
  {
    dstCoef[i] = srcCoef[i];
    sum += (2* srcCoef[i]);
  }

  Int pred = (1<<AF_NUM_BIT_SHIFT) - (sum);
  //center coefficients
  dstCoef[numCoef-1] = srcCoef[numCoef-1] + pred;

}
#endif
//! \}

